/**
 * @file qoco.h
 * @author Govind M. Chari <govindchari1@gmail.com>
 *
 * @section LICENSE
 *
 * Copyright (c) 2024, Govind M. Chari
 * This source code is licensed under the BSD 3-Clause License
 *
 * @section DESCRIPTION
 *
 * This is the file that should be included when using QOCO.
 */

#ifndef QOCO_H
#define QOCO_H

#ifdef __cplusplus
extern "C" {
#endif

#include "cone.h"
#include "definitions.h"
#include "qoco_api.h"
#include "qoco_utils.h"
#include "structs.h"

#ifdef __cplusplus
}
#endif

#endif /* #ifndef QOCO_H */