use crate::solver::implementations::default::DefaultSettings;

/// Solver general core settings are the same as in the default solver.
///
/// Go [here](crate::solver::implementations::default::DefaultSettings)
/// to view the complete list.
///
pub type CoreSettings<T> = DefaultSettings<T>;
